#include "tcpip.h"

/* Generator */
struct generator tcp_gen=init_generator("sig.tcp");

struct proto tcp_p=init_proto2("tcp", tcp_decode, tcp_dprint, NULL, tcp_commit);
struct proto_req tcp_r[]={
	proto_request("ip", 6),
	null_request()
};

int tcp_dprint(struct layer *l, char *buf, int buflen)
{
	char tcpflags[]="FSRPAU12";
	char disp[9];
	int n, x;

	for(n=0,x=1; n<8; n++, x*=2) {
		if ( l->h.tcp->flags.flags & x ){
			disp[n]=tcpflags[n];
		}else{
			disp[n]='*';
		}
	}disp[8]='\0';

	return snprintf(buf, buflen,
		"%u > %u [%s] seq=0x%x ack=0x%x win=%u",
		ntohs(l->h.tcp->sport),
		ntohs(l->h.tcp->dport),
		disp,
		ntohl(l->h.tcp->seq),
		ntohl(l->h.tcp->ack),
		ntohs(l->h.tcp->win));
}

static inline u_int32_t tcp_csum(struct packet *pkt, unsigned int l)
{
	struct pkt_iphdr *iph=pkt->layer[l].h.ip;
	struct pkt_tcphdr *tcph=pkt->layer[l+1].h.tcp;
	struct tcp_phdr ph;
	u_int16_t *tmp;
	u_int32_t sum=0;
	u_int16_t len, csum;
	int i;

	if ( pkt->layer[l].flags&FLAG_IP_DTRUNC )
		return 0;

	len=ntohs(iph->tot_len) - (iph->ihl<<2);

	/* Make pseudo-header */
	ph.sip=iph->saddr;
	ph.dip=iph->daddr;
	ph.zero=0;
	ph.proto=iph->protocol;
	ph.tcp_len=htons(len);

	/* Checksum the pseudo-header */
	tmp=(u_int16_t *)&ph;
	for(i=0; i<6; i++) {
		sum+=tmp[i];
		if(sum & 0x80000000) {
			sum = (sum & 0xffff) + (sum >> 16);
		}
	}

	/* Checksum the header+data */
	tmp=(u_int16_t *)tcph;
	for(i=0; i<len>>1; i++) {
		sum+=tmp[i];
		if(sum & 0x80000000) {
			sum = (sum & 0xffff) + (sum >> 16);
		}
	}

	/* Deal with last byte (if odd number of bytes) */
	if ( len&1 ) {
		union {
			u_int8_t b[2];
			u_int16_t s;
		}f;

		f.b[0]=((u_int8_t *)tcph)[len-1];
		f.b[1]=0;
		sum+=f.s;
	}

	while(sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);

	csum=(~sum) & 0xffff;

	return csum ? 0 : FLAG_TCP_CSUM;
}

void tcp_decode(struct packet *p)
{
	struct layer *l=&p->layer[p->llen];

	/* Check it is at least minimum size */
	if ( l->h.raw+sizeof(struct pkt_tcphdr) > p->end ) {
		return;
	}

	/* Use the doff field to decode the packet */
	if ( (p->layer[p->llen+1].h.raw=
		l->h.raw+(l->h.tcp->doff<<2))
		> p->end ) {
		return;
	}

	p->llen++;

	/* Run IP pre-processors */
	if ( tcp_p.num_active ) {
		int i;
		for(i=0; i<tcp_p.num_active; i++) {
			struct pp_active *pp=&tcp_p.preproc[i];
			pp->process(p, p->llen-1, pp->priv);
		}
	}

	if ( p->llen >= PKT_LAYERS ) return;

	/* State tracking */
	if ( tcp_stateful ) {
		l->flags|=tcp_csum(p, p->llen-2);
		tcpstream_process(p, p->llen-2);
		return;
	}

	p->layer[p->llen++].proto=NULL;
}


/* ===================================================
 * PACKET MATCHING STUFF BEYOND THIS POINT
 * ===================================================
 */

struct sig_node tcp_root;

void tcp_match(struct packet *p, unsigned int l)
{
	detect_set(&cur_alert, &alert_depth);
	detect(tcp_root.child, p, l);
	if ( cur_alert ) alert(&tcp_gen, p, cur_alert);
}

int tcp_commit(struct rule *r)
{
	struct criteria *c;
	struct sig_node *x;
	struct matcher *m;
	struct alert *a;
	unsigned int i;
	int forceflow= (tcp_stateful) ? 1 : 0;

	/* Can't handle this case */
	if ( r->num_criteria==0 ) return 0;

	if ( !(x=calloc(r->num_criteria+1, sizeof(*x))) ) return 0;

	/* Copy the alert data */
	if ( !(a=calloc(1, sizeof(*a))) ) {
		free(x);
		return 0;
	}
	memcpy(a, &r->alert, sizeof(*a));

	/* Build an array of all the nodes we want to add */
	for(c=r->criteria,i=0; c; c=c->next,i++) {
		char *key;

		if ( !strcmp(c->crit, "src") ) {
			key="ip_src";
		}else if ( !strcmp(c->crit, "dst") ) {
			key="ip_dst";
		}else if ( !strcmp(c->crit, "sport") ) {
			key="tcp_sport";
		}else if ( !strcmp(c->crit, "dport") ) {
			key="tcp_dport";
		}else if ( !strcmp(c->crit, "flow") || 
			!strcmp(c->crit, "stateless") ) {
			if ( !tcp_stateful ) {
				/* dont allow flow or stateless if tcpstream
				 * is disabled */
				forceflow=-1;
				i--;
				continue;
			}
			forceflow=0;
			key=c->crit;
		}else key=c->crit;

		if ( !(m=matcher_find(key)) ) {
			mesg(M_ERR,"tcp_commit: cannot find '%s' matcher", key);
			detect_free_sig(x, i);
			return 0;
		}

		x[i].match=m;
		x[i].n=c->negate;
		x[i].cost=m->cost;
		if ( !(x[i].m=m->validate(c->args,
			&x[i].p, c->modifier, &x[i].cost)) ) {
			detect_free_sig(x, i);
			return 0;
		}
	}

	/* Enforce a default policy on flow */
	if ( forceflow==1 ) {
		if ( !(m=matcher_find("flow")) ) {
			mesg(M_ERR,"tcp_commit: cannot find '%s' matcher", "flow");
			detect_free_sig(x, i);
			return 0;
		}

		x[i].match=m;
		x[i].n=0;
		x[i].cost=m->cost;
		if ( !(x[i].m=m->validate(NULL,&x[i].p, NULL, &x[i].cost)) ) {
			detect_free_sig(x, i);
			return 0;
		}
	}

	return detect_add_sig(x, r->num_criteria+forceflow, &tcp_root, a);
}

